// WATER JUGS PROBLEM - Homework 1 for AI - Beycan Kahraman  -  040020337

#include <stdio.h>
#include <stdlib.h>

#define 	BFS 			1 		// Breadth First Search method
#define 	DFS 			2 		// Depth First Search method
#define 	UCS 			3 		// Uniform Cost Search method

#define 	OPEN   			1 		// Node is generated but not expanded yet
#define 	CLOSED 			2 		// Node is already expanded
#define 	NONE 			-1 		// Node has not opened yet

typedef struct node_s{
	int		depth; // BFS & DFS Search Criterion
	int 	water; // UCS Criterion
	int		parent;
	int 	action; // added in order not to calculate the action again
	int		flag;
}node;

void Error(char str[]);
void Search(int strategy, int jugs[], node nodes[]);
void DisplayPath(int jugs[], node nodes[]);

int main()
{
	int 	jugs[6]; 	// { Capacity1, Capacity2, Init1, Init2, Goal1, Goal2 }
	int 	strategy;
	node	*nodes		= NULL;

	printf("WATER JUGS PROBLEM    (Beycan Kahraman - 040020337)\n\n");
	
	printf("Enter capacities of the small jug and big jug: ");
	scanf("%d %d", &jugs[0], &jugs[1]);
	if(jugs[0] < 1 || jugs[1] < 1)
		Error("Capacities of Water Jugs Can Not Be Non-Positive");

	nodes = malloc( sizeof(node)*(jugs[0]+1)*(jugs[1]+1) );
	if(nodes == NULL)
		Error("You Do Not Have Enough Memory to Solve This Problem");
	
    printf("Enter initial states of jugs: ");
    scanf("%d %d", &jugs[2], &jugs[3]);
    if(jugs[2] < 0 || jugs[3] < 0 || jugs[2] > jugs[0] || jugs[3] > jugs[1])
		Error("No Possibility for Water Jugs' Initial Satate");
		
    printf("Enter goal states of jugs: ");
    scanf("%d %d", &jugs[4], &jugs[5]);
    if(jugs[4] < 0 || jugs[5] < 0 || jugs[4] > jugs[0] || jugs[5] > jugs[1])
		Error("No Possibility for Water Jugs' Goal State");
		
	printf("Enter search strategy (1.BFS, 2.DFS, 3.UCS): ");
	scanf("%d", &strategy);
    if(strategy < 1 || strategy > 3)
		Error("Search Strategy Does Not Defined");
	
	Search(strategy, jugs, nodes);
	
	DisplayPath(jugs, nodes);
	
	free( nodes );
	
	system("PAUSE");
	return 0;
}

void Error(char str[]) // prints the errors
{
	printf("\nERROR: %s\n\n", str);
	system("PAUSE");
	exit(-1);
}

void Expand(int strategy, int current, int jugs[], node nodes[])
{
	int jug1 = current%(jugs[0] + 1), jug2 = current/(jugs[0] + 1);
	nodes[current].flag = CLOSED;
	
	// There are 6 types of expanding for this question
	
	// Fill jug1
	if(nodes[jugs[0] + jug2*(jugs[0] + 1)].flag == NONE || 
		(strategy == UCS && nodes[current].water + 0 < nodes[jugs[0] + jug2*(jugs[0] + 1)].water))
	{
		nodes[jugs[0] + jug2*(jugs[0] + 1)].flag = OPEN;
		nodes[jugs[0] + jug2*(jugs[0] + 1)].action = 2;
		nodes[jugs[0] + jug2*(jugs[0] + 1)].parent = current;
		nodes[jugs[0] + jug2*(jugs[0] + 1)].water = nodes[current].water + 0;
		nodes[jugs[0] + jug2*(jugs[0] + 1)].depth = nodes[current].depth + 1;
	}
	
	// Fill Jug2
	if(nodes[jug1 + jugs[1]*(jugs[0] + 1)].flag == NONE || 
		(strategy == UCS && nodes[current].water + 0 < nodes[jug1 + jugs[1]*(jugs[0] + 1)].water))
	{
		nodes[jug1 + jugs[1]*(jugs[0] + 1)].flag = OPEN;
		nodes[jug1 + jugs[1]*(jugs[0] + 1)].action = 1;
		nodes[jug1 + jugs[1]*(jugs[0] + 1)].parent = current;
		nodes[jug1 + jugs[1]*(jugs[0] + 1)].water = nodes[current].water + 0;
		nodes[jug1 + jugs[1]*(jugs[0] + 1)].depth = nodes[current].depth + 1;
	}
	
	// Empty Jug1
	if(nodes[0 + jug2*(jugs[0] + 1)].flag == NONE || 
		(strategy == UCS && nodes[current].water + jug1 < nodes[0 + jug2*(jugs[0] + 1)].water))
	{
		nodes[0 + jug2*(jugs[0] + 1)].flag = OPEN;
		nodes[0 + jug2*(jugs[0] + 1)].action = 4;
		nodes[0 + jug2*(jugs[0] + 1)].parent = current;
		nodes[0 + jug2*(jugs[0] + 1)].water = nodes[current].water + jug1;
		nodes[0 + jug2*(jugs[0] + 1)].depth = nodes[current].depth + 1;
	}
	
	// Empty Jug2
	if(nodes[jug1 + 0*(jugs[0] + 1)].flag == NONE || 
		(strategy == UCS && nodes[current].water + jug2 < nodes[jug1 + 0*(jugs[0] + 1)].water))
	{
		nodes[jug1 + 0*(jugs[0] + 1)].flag = OPEN;
		nodes[jug1 + 0*(jugs[0] + 1)].action = 3;
		nodes[jug1 + 0*(jugs[0] + 1)].parent = current;
		nodes[jug1 + 0*(jugs[0] + 1)].water = nodes[current].water + jug2;
		nodes[jug1 + 0*(jugs[1] + 1)].depth = nodes[current].depth + 1;
	}
	
	// Jug1 --> Jug2
	if(jug1 > 0 && jug2 < jugs[1])
		if(jugs[1] - jug2 > jug1){
			if(nodes[0 + (jug2 + jug1)*(jugs[0] + 1)].flag == NONE || 
		        (strategy == UCS && nodes[current].water + 0 < nodes[0 + (jug2 + jug1)*(jugs[0] + 1)].water))
			{
				nodes[0 + (jug2 + jug1)*(jugs[0] + 1)].flag = OPEN;
				nodes[0 + (jug2 + jug1)*(jugs[0] + 1)].action = 5;
				nodes[0 + (jug2 + jug1)*(jugs[0] + 1)].parent = current;
				nodes[0 + (jug2 + jug1)*(jugs[0] + 1)].water = nodes[current].water + 0;
				nodes[0 + (jug2 + jug1)*(jugs[0] + 1)].depth = nodes[current].depth + 1;
			}
		}else{
			if(nodes[(jug1 - jugs[1] + jug2) + jugs[1]*(jugs[0] + 1)].flag == NONE || 
		        (strategy == UCS && nodes[current].water + 0 < nodes[(jug1 - jugs[1] + jug2) + jugs[1]*(jugs[0] + 1)].water))
			{
				nodes[(jug1 - jugs[1] + jug2) + jugs[1]*(jugs[0] + 1)].flag = OPEN;
				nodes[(jug1 - jugs[1] + jug2) + jugs[1]*(jugs[0] + 1)].action = 5;
				nodes[(jug1 - jugs[1] + jug2) + jugs[1]*(jugs[0] + 1)].parent = current;
				nodes[(jug1 - jugs[1] + jug2) + jugs[1]*(jugs[0] + 1)].water = nodes[current].water + 0;
				nodes[(jug1 - jugs[1] + jug2) + jugs[1]*(jugs[0] + 1)].depth = nodes[current].depth + 1;
			}
		}
	
	// Jug2 --> Jug1
	if(jug2 > 0 && jug1 < jugs[0]){
		if(jugs[0] - jug1 > jug2){
			if(nodes[(jug1 + jug2) + 0*(jugs[0] + 1)].flag == NONE || 
		        (strategy == UCS && nodes[current].water + 0 < nodes[(jug1 + jug2) + 0*(jugs[0] + 1)].water))
			{
				nodes[(jug1 + jug2) + 0*(jugs[0] + 1)].flag = OPEN;
				nodes[(jug1 + jug2) + 0*(jugs[0] + 1)].action = 6;
				nodes[(jug1 + jug2) + 0*(jugs[0] + 1)].parent = current;
				nodes[(jug1 + jug2) + 0*(jugs[0] + 1)].water = nodes[current].water + 0;
				nodes[(jug1 + jug2) + 0*(jugs[0] + 1)].depth = nodes[current].depth + 1;
			}
		}else{
			if(nodes[jugs[0] + (jug2 + jug1 - jugs[0])*(jugs[0] + 1)].flag == NONE || 
                (strategy == UCS && nodes[current].water + 0 < nodes[jugs[0] + (jug2 + jug1 - jugs[0])*(jugs[0] + 1)].water))
			{
				nodes[jugs[0] + (jug2 + jug1 - jugs[0])*(jugs[0] + 1)].flag = OPEN;
				nodes[jugs[0] + (jug2 + jug1 - jugs[0])*(jugs[0] + 1)].action = 6;
				nodes[jugs[0] + (jug2 + jug1 - jugs[0])*(jugs[0] + 1)].parent = current;
				nodes[jugs[0] + (jug2 + jug1 - jugs[0])*(jugs[0] + 1)].water = nodes[current].water + 0;
				nodes[jugs[0] + (jug2 + jug1 - jugs[0])*(jugs[0] + 1)].depth = nodes[current].depth + 1;
			}
		}
	}
}

void Search(int strategy, int jugs[], node nodes[])
{
	int i, chosen, searchdepth, count = 1;
	int currentJug = jugs[2] + jugs[3]*(jugs[0]+1);
	int goalJug 	= jugs[4] + jugs[5]*(jugs[0]+1);
	
	for(i=0; i<(jugs[0]+1)*(jugs[1]+1); ++i)
	{
		nodes[i].depth 		= 0;
		nodes[i].flag 		= NONE;
		nodes[i].water 		= 0;
		nodes[i].action 	= 0;
		nodes[i].parent 	= 0;
	}
	
	while(currentJug != goalJug)
	{
		chosen = -1;
		++count;
		searchdepth = strategy == DFS ? 0 : 0x79999999;
		
		Expand(strategy, currentJug, jugs, nodes);
		
		for(i=0; i<(jugs[0]+1)*(jugs[1]+1); ++i)
		{
			if (nodes[i].flag != OPEN)
          		continue;
          
			switch(strategy)
			{
				case BFS: 
					if(nodes[i].depth < searchdepth)
					{
						searchdepth = nodes[i].depth;
						chosen = i;
					}
					break;
				case DFS: 
					if(nodes[i].depth > searchdepth)
					{
						searchdepth = nodes[i].depth;
						chosen = i;
					}
					break;
				case UCS: 
					if(nodes[i].water < searchdepth)
					{
						searchdepth = nodes[i].water;
						chosen = i;
					}
					break;
					
				default: Error("Strategy Could Not Be Found"); break;
			}
		}
		
		if(chosen < 0)
		{
			printf ("\nThere is no solution for these inputs!..\n\n");
			system("PAUSE");
			exit(-1);
		}
		
		currentJug = chosen;
	}
	
	printf("\nCalculated in %d Steps  (Order of Node Choosing:  jug1 + %d*jug2)\n", count, jugs[0]+1);
	// Order of chosing is the scanning method of the nodes in the array
	// Search in : 00 - 10 - 20 - 30 - 01 - 11 - 21 - 31 - 02 - 12 - ...  Order
}

void DisplayPath(int jugs[], node nodes[])
{
	int 	toBePrinted = 0;
	int 	count = 0;
	int		*path;
	int		*actions;
	
	printf ("\nSOLUTION PATH:\n");
	
	path = malloc( sizeof(int)*(jugs[0]+1)*(jugs[1]+1) );
	actions = malloc( sizeof(int)*(jugs[0]+1)*(jugs[1]+1) );
	if(actions == NULL)
		Error("You Do Not Have Enough Memory to Solve This Problem");
	
	path[0] = jugs[4] + jugs[5]*(jugs[0]+1);    // Select the goal state
	actions[0] = nodes[jugs[4] + jugs[5]*(jugs[0]+1)].action;
	
	// Find the way to starting state
	while(path[toBePrinted] != jugs[2] + jugs[3]*(jugs[0]+1))
	{
		path[toBePrinted+1] = nodes[path[toBePrinted]].parent;
		actions[toBePrinted] = nodes[path[toBePrinted]].action;
		++toBePrinted;
	}
	
	actions[toBePrinted] = 0;
	
	// print the actions and path in reverse order
	while(toBePrinted >= 0)
	{
		printf ("STEP=%d STATE=(%d,%d) WASTED=%d ACTION=", 
			count, path[toBePrinted]%(jugs[0]+1), path[toBePrinted]/(jugs[0]+1), nodes[path[toBePrinted]].water);
			
		switch(actions[toBePrinted])
		{
			case 0: printf("  (No Action)\n"); break;
			case 1: printf("1 (Fill %dL. Jug)\n", jugs[1]); break;
			case 2: printf("2 (Fill %dL. Jug)\n", jugs[0]); break;
			case 3: printf("3 (Empty %dL. Jug)\n", jugs[1]); break;
			case 4: printf("4 (Empty %dL. Jug)\n", jugs[0]); break;
			case 5: printf("5 (Pour From %dL. Jug to %dL. Jug)\n", jugs[0], jugs[1]); break;
			case 6: printf("6 (Pour From %dL. Jug to %dL. Jug)\n", jugs[1], jugs[0]); break;
			default: printf("%d (ERROR)\n", actions[toBePrinted]); break;
		}
		
		++count;
		--toBePrinted;
	}
	
	free( actions );
	free( path );
	printf ("Searching finished\n\n");
}
